`include "vc-muxes.v"

// Slightly modified from the DE2 default design example
module VGA_Controller
(
 // This clock should be the VGA pixel clock, which is 25 MHz for a
 // resolution of 640 x 480 at a 60 Hz refresh rate
 input         clk,
 input         reset,

 // Host Side
 output [18:0] memory_address,
 input [7:0]   memory_data,

 // VGA Side
 output [7:0]  vga_r,
 output [7:0]  vga_g,
 output [7:0]  vga_b,
 output        vga_h_sync,
 output        vga_v_sync,
 output        vga_sync,
 output        vga_blank
);

  //  Horizontal Parameter  ( Pixel )
  localparam h_sync_a = 10'd96; // 3.8 us * 25.175 MHz
  localparam h_sync_b = 10'd48;
  localparam h_sync_c = 10'd640;
  localparam h_sync_d = 10'd16;

  localparam h_sync_ab   = h_sync_a + h_sync_b;
  localparam h_sync_abc  = h_sync_a + h_sync_b + h_sync_c;
  localparam h_sync_abcd = h_sync_a + h_sync_b + h_sync_c + h_sync_d;

  //  Vertical Parameter    ( Line )
  localparam v_sync_a = 10'd2;
  localparam v_sync_b = 10'd33;
  localparam v_sync_c = 10'd480;
  localparam v_sync_d = 10'd10;

  localparam v_sync_ab   = v_sync_a + v_sync_b;
  localparam v_sync_abc  = v_sync_a + v_sync_b + v_sync_c;
  localparam v_sync_abcd = v_sync_a + v_sync_b + v_sync_c + v_sync_d;

  //----------------------------------------------------------------------
  // VGA mux
  //----------------------------------------------------------------------
  reg             vga_mux_sel;
  vc_Mux2 #(24) vga_mux
  (
   .in0 (24'd0),
   .in1 ({ memory_data, memory_data, memory_data }),
   .sel (vga_mux_sel),
   .out ({ vga_r, vga_g, vga_b })
  );

  //----------------------------------------------------------------------
  // Pixel counter
  //----------------------------------------------------------------------
  reg [9:0]       h_count;
  always @ ( posedge clk ) begin
    if ( reset ) begin
      h_count <= 10'b1;
    end else begin
      if ( h_count == h_sync_abcd ) // Reset back
        h_count <= 10'b1;
      else
        h_count <= h_count + 10'b1;
    end
  end

  //----------------------------------------------------------------------
  // Line counter
  //----------------------------------------------------------------------
  reg [9:0]       v_count;
  wire            h_max = (h_count == h_sync_abcd);
  always @ ( posedge clk ) begin
    if ( reset ) begin
      v_count <= 10'b1;
    end else begin
      if ( h_max ) begin
        if ( v_count == v_sync_abcd )
          v_count <= 10'b1;
        else
          v_count <= v_count + 10'b1;
      end else begin
        v_count <= v_count;
      end
    end
  end


  //----------------------------------------------------------------------
  // Outputs
  //----------------------------------------------------------------------

  assign vga_v_sync = ~(v_count <= v_sync_a);
  assign vga_h_sync = ~(h_count <= h_sync_a);
  assign vga_blank  = vga_h_sync && vga_v_sync;
  assign vga_sync   = 1'b0;

  always @ ( * ) begin
    if ( v_sync_ab < v_count && v_count <= v_sync_abc &&
         h_sync_ab < h_count && h_count <= h_sync_abc ) begin
      vga_mux_sel <= 1'b1;
    end else begin
      vga_mux_sel <= 1'b0;
    end
  end

  // Addressing memory
  reg [8:0] row;
  reg [9:0] col;
  wire [9:0] row_computed = v_count - v_sync_ab - 10'b1;
  always @ ( * ) begin
    if ( v_sync_ab < v_count && v_count <= v_sync_abc &&
         h_sync_ab < h_count && h_count <= h_sync_abc ) begin
      row <= row_computed[8:0];
      col <= h_count - h_sync_ab;
    end else begin
      row <= 9'b0;
      col <= 9'b0;
    end
  end
  assign memory_address = { row, col };

endmodule
