//=========================================================================
// Verilog Components: ClockDivider
//=========================================================================
// Clock divider takes any clock with frequency f as input and outputs a
// clock with frequency f/divisor
// Supports both odd and even integral divisors
// Odd clock division based on http://www.mikrocontroller.net/attachment/177198/Clock_Dividers_Made_Easy.pdf

`ifndef VC_CLOCKDIVIDER_V
`define VC_CLOCKDIVIDER_V

`include "vc-Counter.v"
`include "vc-TFlipFlop.v"


module vc_ClockDivider
#(
  parameter divisor = 16
)(
  input  reset,
  input  input_clk,
  output output_clk
);

  `ifdef XILINX_SYNTHESIS
  `include "vc-fpga-xclog2.v"
  localparam divisor_bitwidth = xclog2(divisor);
  `else
  localparam divisor_bitwidth = $clog2(divisor);
  `endif
  localparam neg_tff_t_value = (divisor-1)/2;

  //----------------------------------------------------------------------
  // For odd clock divisions:
  //----------------------------------------------------------------------

  //----------------------------------------------------------------------
  // Counter
  //----------------------------------------------------------------------

  localparam odd_counter_max = divisor - 1;
  wire [divisor_bitwidth-1:0] count_odd;
  wire                        count_odd_is_max;
  localparam [divisor_bitwidth-1:0] count_reset_value = 'h0;
  vc_Counter
  #(
    .p_count_nbits       (divisor_bitwidth),
    .p_count_reset_value (count_reset_value),
    .p_count_max_value   (divisor-1)
  )
  counter_odd
  (
   .clk           (input_clk),
   .reset         (reset || count_odd_is_max),
   .increment     (1'b1),
   .decrement     (1'b0),
   .count         (count_odd),
   .count_is_zero (),
   .count_is_max  (count_odd_is_max)
  );

  //----------------------------------------------------------------------
  // T-flip-flop (posedge)
  //----------------------------------------------------------------------

  wire                       pos_tff_t = count_odd_is_max;
  wire                       pos_tff_out;
  vc_TFlipFlop
  #(
    .reset_value (1'b1)
  )
  pos_tff
  (
    .clk   (input_clk),
    .reset (reset),
    .t     (pos_tff_t),
    .q     (pos_tff_out)
  );

  //----------------------------------------------------------------------
  // T-flip-flop (negedge)
  //----------------------------------------------------------------------

  wire                       neg_tff_t = (count_odd == neg_tff_t_value);
  wire                       neg_tff_out;
  vc_TFlipFlop
  #(
    .reset_value (1'b1)
  )
  neg_tff
  (
    .clk   (~input_clk),
    .reset (reset),
    .t     (neg_tff_t),
    .q     (neg_tff_out)
  );

  wire                       odd_clock_division_out = ~(pos_tff_out ^ neg_tff_out);

  //----------------------------------------------------------------------
  // For even clock divisions:
  //----------------------------------------------------------------------

  //----------------------------------------------------------------------
  // Counter
  //----------------------------------------------------------------------

  localparam counter_even_max = divisor/2-1;
  wire                        count_even_is_max;
  vc_Counter
  #(
    .p_count_nbits       (divisor_bitwidth),
    .p_count_reset_value (count_reset_value),
    .p_count_max_value   (counter_even_max)
  )
  counter_even
  (
   .clk           (input_clk),
   .reset         (reset || count_even_is_max),
   .increment     (1'b1),
   .decrement     (1'b0),
   .count         (),
   .count_is_zero (),
   .count_is_max  (count_even_is_max)
  );

  wire                       even_tff_t = count_even_is_max;
  wire                       even_tff_out;
  vc_TFlipFlop
  #(
    .reset_value (1'b1)
  )
  even_tff
  (
    .clk   (input_clk),
    .reset (reset),
    .t     (even_tff_t),
    .q     (even_tff_out)
  );

  wire                       even_clock_division_out = even_tff_out;

  //----------------------------------------------------------------------
  // Assign outputs
  //----------------------------------------------------------------------

  // If the divisor is even, use the simpler divider
  // If the divisor is odd, use the more complicated divider
  assign output_clk = divisor[0] ? odd_clock_division_out : even_clock_division_out;

endmodule

`endif
